package com.bcxin.tenant.data.etc.tasks.components;
import com.bcxin.event.core.JsonProvider;
import com.bcxin.event.core.JsonProviderImpl;
import com.bcxin.event.core.exceptions.SkipRetryEventException;
import com.bcxin.event.job.core.domain.CacheProvider;
import com.bcxin.event.job.core.domain.dtos.RedisConfig;
import com.bcxin.event.job.core.domain.impls.CacheProviderImpl;
import org.apache.commons.lang3.ThreadUtils;
import com.bcxin.event.core.KafkaConstants;
import com.bcxin.tenant.data.etc.tasks.dtos.DtqRecordDto;
import org.apache.commons.lang3.exception.ExceptionUtils;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.connector.jdbc.JdbcConnectionOptions;
import org.apache.flink.connector.jdbc.JdbcExecutionOptions;
import org.apache.flink.connector.jdbc.internal.JdbcOutputFormat;
import org.apache.flink.connector.jdbc.internal.connection.SimpleJdbcConnectionProvider;
import org.apache.flink.util.concurrent.ExecutorThreadFactory;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.event.Level;
import org.springframework.util.CollectionUtils;
import org.springframework.util.StopWatch;
import org.springframework.util.StringUtils;

import javax.annotation.Nonnull;
import javax.sql.DataSource;
import java.io.IOException;
import java.io.Serializable;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.time.Duration;
import java.time.temporal.ChronoUnit;
import java.util.*;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

public class CustomJdbcOutputFormat extends JdbcOutputFormat implements Serializable,Runnable {
    private static final Logger logger = LoggerFactory.getLogger(CustomJdbcOutputFormat.class);

    private final JdbcConnectionOptions jdbcConnectionOptions;
    /**
     * 通过Open的时候来初始化DataSource
     */
    private volatile DataSource dataSource;
    private final List<Object> batch;
    private final CustomJdbcAcceptPreparedStatementParameter acceptPreparedStatementParameter;
    private final String sql;
    private final JdbcExecutionOptions executionOptions;
    private transient ScheduledExecutorService scheduler;
    private final String bootstrapServer;
    private final RedisConfig redisConfig;

    private final Collection<String> batchOrConditionExpress;

    private transient ScheduledFuture<?> scheduledFuture;
    private transient KafkaOutputFormat kafkaOutputFormat;




    public CustomJdbcOutputFormat(
            Collection<String> batchOrConditionExpress,
            @Nonnull JdbcConnectionOptions jdbcConnectionOptions,
                                  @Nonnull JdbcExecutionOptions executionOptions,
                                  @Nonnull StatementExecutorFactory statementExecutorFactory,
                                  @Nonnull String sql,
                                  @Nonnull String bootstrapServer,
                                  RedisConfig redisConfig,
                                  CustomJdbcAcceptPreparedStatementParameter acceptPreparedStatementParameter,
                                  @Nonnull RecordExtractor recordExtractor) {
        super(new SimpleJdbcConnectionProvider(jdbcConnectionOptions), executionOptions, statementExecutorFactory, recordExtractor);
        this.batchOrConditionExpress = batchOrConditionExpress;
        this.jdbcConnectionOptions = jdbcConnectionOptions;
        this.batch = new ArrayList<>();
        this.sql = sql;
        this.acceptPreparedStatementParameter = acceptPreparedStatementParameter;
        this.executionOptions = executionOptions;
        this.bootstrapServer = bootstrapServer;
        this.redisConfig = redisConfig;

        Runtime.getRuntime().addShutdownHook(new Thread(this));
    }

    @Override
    public void configure(Configuration parameters) {
        super.configure(parameters);
    }


    private transient CacheProvider cacheProvider;
    @Override
    public void open(int taskNumber, int numTasks) throws IOException {
        //super.open(taskNumber, numTasks);
        this.dataSource = DataSourceUtil.getDataSource(jdbcConnectionOptions.getDriverName(),
                jdbcConnectionOptions.getDbURL(),
                jdbcConnectionOptions.getUsername().get(),
                jdbcConnectionOptions.getPassword().get()
        );

        this.kafkaOutputFormat = new KafkaOutputFormat(
                this.bootstrapServer,
                KafkaConstants.getDtqTopic(KafkaConstants.DTQ_JDBC_CONSUMER_TOPIC),
                100, 5_000);

        this.kafkaOutputFormat.open(taskNumber, numTasks);
        if (executionOptions.getBatchIntervalMs() != 0 && executionOptions.getBatchSize() != 1) {
            this.scheduler =
                    Executors.newScheduledThreadPool(
                            1, new ExecutorThreadFactory("jdbc-upsert-output-format"));
            this.scheduledFuture =
                    this.scheduler.scheduleWithFixedDelay(
                            () -> {
                                synchronized (CustomJdbcOutputFormat.class) {
                                    try {
                                        flush();
                                    } catch (Exception e) {
                                        e.printStackTrace();
                                    }
                                }
                            },
                            executionOptions.getBatchIntervalMs(),
                            executionOptions.getBatchIntervalMs(),
                            TimeUnit.MILLISECONDS);
        }

        cacheProvider = new CacheProviderImpl(redisConfig);
    }

    @Override
    public synchronized void flush() throws IOException {
        super.flush();
    }

    @Override
    public synchronized void close() {
        if (this.scheduledFuture != null) {
            this.scheduledFuture.cancel(true);
        }

        if (this.cacheProvider != null) {
            this.cacheProvider.close();
        }
    }

    @Override
    public Connection getConnection() {
        //return super.getConnection();
        return null;
    }

    @Override
    public void updateExecutor(boolean reconnect) throws SQLException, ClassNotFoundException {
        /*
        super.updateExecutor(reconnect);
         */
    }


    /**
     * 该方法是线程安全的
     * @param original
     * @param extracted
     * @throws SQLException
     */
    @Override
    protected void addToBatch(Object original, Object extracted) throws SQLException {
        this.batch.add(extracted);
    }

    @Override
    protected void attemptFlush() throws SQLException {
        if (!batch.isEmpty()) {
            StringBuilder cost = new StringBuilder();
            cost.append(String.format("v3-当前batch的数据量=%s;", batch.size()));
            StopWatch stopWatch = new StopWatch();
            Collection<BinlogRawValue> matchRawValues = new ArrayList<>();
            JsonProvider jsonProvider = new JsonProviderImpl();
            String lasExpression = null;
            for (Object r : batch) {
                BinlogRawValue rawValue = (BinlogRawValue) r;
                boolean isMatchExecuteExpress = rawValue.isMatchCondition(jsonProvider, batchOrConditionExpress);
                if (isMatchExecuteExpress) {
                    matchRawValues.add(rawValue);
                } else {
                    lasExpression = rawValue.getConditionExecuteResult();
                }
            }

            stopWatch.start();
            try {
                if (!matchRawValues.isEmpty()) {
                    cost.append(String.format("开始JDBC操作[%s/%s]:", matchRawValues.size(), batch.size()));
                    try (Connection connection = this.dataSource.getConnection()) {
                        connection.setTransactionIsolation(Connection.TRANSACTION_READ_COMMITTED);
                        /**
                         * setCommit的话,会造成Deadlock found when trying to get lock; try restarting transaction
                         * 正在错误信息
                         */
                        String trace = executeSql(connection, matchRawValues);
                        cost.append(trace);
                    }
                } else {
                    cost.append(String.format("总共有%s/%s条数据不符合condition条件[lasExpression=%s];", batch.size() - matchRawValues.size(), batch.size(), lasExpression));
                }
            }
            catch (Exception ex) {
                logger.error(cost.toString(), ex);
            }
            finally {
                try {
                    stopWatch.stop();
                    cost.append(String.format("总共耗时=%s ms", stopWatch.getTotalTimeMillis()));
                } finally {
                    logger.info(cost.toString());
                }
            }

            batch.clear();
        }
    }

    private String executeSql(Connection connection,
                            Collection<BinlogRawValue> valueWrappers) throws SQLException {
        boolean isExecutable = false;
        StringBuilder sb = new StringBuilder();
        try (PreparedStatement statement = connection.prepareStatement(this.sql)) {
            StopWatch topStopWatch = new StopWatch();
            topStopWatch.start("准备参数信息");
            sb.append(String.format("执行的表达式:%s;",
                    (CollectionUtils.isEmpty(batchOrConditionExpress) ? "NULL" : batchOrConditionExpress.stream().collect(Collectors.joining(";"))))
            );

            StringBuilder paramSb = new StringBuilder();
            for (BinlogRawValue o : valueWrappers) {
                try {
                    StopWatch stopWatch = new StopWatch();
                    stopWatch.start("截取对象的值信息");
                    acceptPreparedStatementParameter.accept(statement, o, cacheProvider);
                    stopWatch.stop();
                    statement.addBatch();
                    isExecutable = true;
                    paramSb.append(String.format("%s,", o.getReadyPkId()));
                } catch (Exception ex) {
                    try {
                        String dbName = (String) o.getReadyParameter("source.db");
                        String tableName = (String) o.getReadyParameter("source.table");
                        String id = String.valueOf(o.getReadyPkId());
                        logger.error("跟踪：数据id={}，表名={},添加失败",id,tableName);
                        this.kafkaOutputFormat.writeRecord(DtqRecordDto.create(dbName, tableName, id, new String(o.getValue())));
                        sb.append(String.format("可忽略: 跳过异常并加入死信队列: 无效的jdbc参数信息; 导致该数据无法正常更新到db:%s", ex.toString()));
                    } catch (Exception ee) {
                        sb.append(String.format("消息(%s)推送到死信队列发生失败:%s", new String(o.getValue()), ee.toString()));
                    }
                }
            }

            topStopWatch.stop();
            sb.append(String.format("参数(%s)耗时:%s ms", paramSb, topStopWatch.getTotalTimeMillis()));

            if (isExecutable) {
                boolean isSuccess = true;
                StringBuilder error = new StringBuilder();
                for (int index = 0; index < 10; index++) {
                    try {
                        StopWatch execStopWatch = new StopWatch();
                        execStopWatch.start("开始执行存储过程");
                        int[] result = statement.executeBatch();
                        isSuccess = true;

                        execStopWatch.stop();

                        sb.append(String.format("(执行完毕[数量=%s]:)", valueWrappers.size()));
                        sb.append(String.format("存储过程总共耗时=%s ms", execStopWatch.getTotalTimeMillis()));

                        if (result != null) {
                            long affectedCount = Arrays.stream(result).filter(ix -> ix == 1).count();
                            sb.append(String.format("其中总共有%s条受影响;", affectedCount));
                        }
                        if(valueWrappers.size() > 0){
                            try{
                                BinlogRawValue binlogRawValue = valueWrappers.stream().findFirst().orElse(null);
                                if(binlogRawValue != null){
                                    String tableName = (String) binlogRawValue.getReadyParameter("source.table");
                                    //if("tenant_employees".equals(tableName)){
                                        logger.error("跟踪v5={}：表名={},statement={}执行成功",sb,tableName,statement);
                                    //}
                                }
                            }catch (Exception ex){
                                logger.error("跟踪v5执行异常={} statement={}",sb, statement, ex);
                            }
                        }
                        break;
                    } catch (Exception ex) {
                        logger.error("执行失败的statement={}，exception={},尝试次数={}",statement,ex,index);
                        if (ex instanceof SQLException && ex.toString().contains("java.sql.SQLException: No operations allowed after statement closed.")) {
                            throw ex;
                        }

                        if (ExceptionUtils.getStackTrace(ex).contains("Incorrect")) {
                            logger.error("【不进行重试-v2】执行({})数据库发生异常(参数异常):{}", this.sql, ex);
                            throw new SkipRetryEventException(ex.getMessage(), ex);
                        }

                        isSuccess = false;
                        if (index < 10) {
                            error.append(String.format("可暂时忽略(当前重试索引次数=%s): 将进行10次的重试 %s", index, ex));

                            try {
                                ThreadUtils.sleep(Duration.of(5, ChronoUnit.MILLIS));
                            } catch (Exception ee) {
                            }
                        } else {
                            isSuccess = false;
                            ex.printStackTrace();
                            logger.error("jdbc脚本({}) 重试({}次数)之后还是执行异常：{}", this.sql, index, ex);
                            throw ex;
                        }
                    } finally {
                        if (isSuccess) {
                            if (error.length() > 0) {
                                sb.append(String.format("重试后(当前index=%s)-最终执行成功: jdbc脚本(%s) 重试时错误信息如下:%s", index, this.sql, error));
                            } else {
                                sb.append(String.format("归集执行成功: sql=%s;", this.sql));
                            }
                        }
                    }
                }
            }
        }

        return sb.toString();
    }

    @Override
    public void run() {
        boolean flag = false;
        try {
            this.flush();

            flag = true;
        } catch (IOException e) {
            throw new RuntimeException(e);
        } finally {
            StringBuilder sb = new StringBuilder();
            if (batch == null) {
                sb.append("暂无数据");
            } else {
                sb.append(String.format("批量数据=%s",
                        batch.stream().filter(ii -> ii != null).map(ii -> String.valueOf(ii))
                                .collect(Collectors.joining(";"))));
            }

            logger.error("系统出现非预期的行为导致异常: 可能丢失的数据为:{}, 数据大小={}", flag, sb);
        }
    }
}
